from spyt.dependency_utils import require_yt_client
from spyt.enabler import SpytEnablers

require_yt_client()

from yt.wrapper.run_operation_commands import run_operation  # noqa: E402
from .conf import read_global_conf, read_remote_conf  # noqa: E402
from .spec import build_spark_connect_server_spec  # noqa: E402
from .utils import parse_bool  # noqa: E402
from .version import __scala_version__ as spyt_version  # noqa: E402


def start_connect_server(client, spark_conf: dict = {}, enablers: SpytEnablers = None, prefer_ipv6: bool = False,
                         driver_memory: str = "1G", num_executors: int = 1, executor_cores: int = 1,
                         executor_memory: str = "2G", pool: str = None, grpc_port_start: int = 27080,
                         java_home: str = None):
    global_conf = read_global_conf(client=client)
    version_config = read_remote_conf(global_conf, spyt_version, client)
    java_home = java_home or version_config.get('default_cluster_java_home')

    enable_squashfs = parse_bool(spark_conf.get("spark.ytsaurus.squashfs.enabled"))
    enablers = enablers or SpytEnablers(enable_squashfs=enable_squashfs)
    enablers.apply_config(version_config)

    spec = build_spark_connect_server_spec(client, version_config, spark_conf, enablers, java_home,
                                           driver_memory, num_executors, executor_cores, executor_memory,
                                           prefer_ipv6, pool, grpc_port_start)
    return run_operation(spec, sync=False, client=client)



from pyspark.sql.column import _to_java_column, Column
from pyspark.sql import DataFrame
from pyspark.sql.functions import lit
from pyspark.sql import SparkSession


def join_with_hot_key_null(right, key, joinType, condition=None):
    if condition is None:
        condition = lit(True)

    def join(left):
        java_condition = _to_java_column(condition)
        return DataFrame(
            left._sc._jvm.tech.ytsaurus.spyt.common.utils.DataFrameUtils.joinWithHotKeyNull(
                left._jdf, right._jdf, key, joinType, java_condition
            ),
            left.sql_ctx
        )
    return join


def get_top(schema, top_col_names, select_col_names):
    spark = SparkSession.builder.getOrCreate()
    jschema = spark._jsparkSession.parseDataType(schema.json())
    jc = spark._jvm.tech.ytsaurus.spyt.common.utils.TopUdaf.top(jschema, top_col_names, select_col_names)
    return Column(jc)


def col_to_yson(df, name, new_name=None, skip_nulls=True):
    new_name = new_name or name
    jdf = df._sc._jvm.tech.ytsaurus.spyt.PythonUtils.serializeColumnToYson(df._jdf, name, new_name, skip_nulls)
    return DataFrame(jdf, df.sql_ctx)



import argparse
import logging
import os
import re
import subprocess
from packaging.version import Version
from pyspark import __version__ as spark_version

from spyt.dependency_utils import require_yt_client
require_yt_client()

from yt.wrapper import YPath  # noqa: E402
from yt.wrapper.cypress_commands import list as yt_list, create, exists  # noqa: E402
from yt.wrapper.errors import YtHttpResponseError  # noqa: E402
from yt.wrapper.http_helpers import get_user_name, get_token  # noqa: E402
from yt.wrapper.operation_commands import get_operation_url  # noqa: E402
from yt.yson.convert import yson_to_json  # noqa: E402
from .arcadia import checked_extract_spark, checked_extract_spyt  # noqa: E402

logger = logging.getLogger(__name__)


class SparkCluster(object):
    def __init__(self, master_endpoint, master_web_ui_url, master_web_ui_schemed_url, master_rest_endpoint, operation_id, shs_url,
                 livy_url, spark_cluster_version, children_operation_ids):
        self.master_endpoint = master_endpoint
        self.master_web_ui_url = master_web_ui_url
        self.master_web_ui_schemed_url = master_web_ui_schemed_url
        self.master_rest_endpoint = master_rest_endpoint
        self.operation_id = operation_id
        self.shs_url = shs_url
        self.livy_url = livy_url
        self.spark_cluster_version = spark_cluster_version
        self.children_operation_ids = children_operation_ids

    def master_webui(self):
        return self.master_web_ui_schemed_url or "http://{}/".format(self.master_web_ui_url)

    def operation_url(self, client=None):
        return get_operation_url(self.operation_id, client=client)

    def children_operation_urls(self, client=None):
        return [get_operation_url(id, client=client) for id in self.children_operation_ids]


class SparkDiscovery(object):
    def __init__(self, discovery_path=None):
        discovery_path = discovery_path or os.getenv("SPARK_BASE_DISCOVERY_PATH")
        self.base_discovery_path = YPath(discovery_path)

    @staticmethod
    def getAll(path, client=None):
        try:
            return yt_list(path, client=client)
        except YtHttpResponseError as e:
            logging.debug("Failed to get path {}, message: {}".format(path, e.message))
            raise e

    @staticmethod
    def get(path, client=None):
        all = SparkDiscovery.getAll(path, client)
        if not all:
            return None
        else:
            return all[0]

    @staticmethod
    def getOption(path, client=None):
        try:
            return SparkDiscovery.get(path, client)
        except YtHttpResponseError:
            return None

    @staticmethod
    def getOptions(path, client=None):
        try:
            return SparkDiscovery.getAll(path, client)
        except YtHttpResponseError:
            return []

    def create(self, client):
        create("map_node", self.discovery(), recursive=True, ignore_existing=True, client=client)
        create("map_node", self.logs(), recursive=True, ignore_existing=True, client=client)

    def discovery(self):
        return self.base_discovery_path.join("discovery")

    def operation(self):
        return self.discovery().join("operation")

    def children_operations(self):
        return self.discovery().join("children_operations")

    def logs(self):
        return self.base_discovery_path.join("logs")

    def event_log(self):
        return self.logs().join("event_log")

    def event_log_table(self):
        return self.logs().join("event_log_table")

    def worker_log(self):
        return self.logs().join("worker_log")

    def master_spark(self):
        return self.discovery().join("spark_address")

    def master_webui(self):
        return self.discovery().join("webui")

    def master_rest(self):
        return self.discovery().join("rest")

    def shs(self):
        return self.discovery().join("shs")

    def livy(self):
        return self.discovery().join("livy")

    def stderr(self):
        return self.logs().join("stderr")

    def spark_cluster_version(self):
        return self.discovery().join("version")

    def set_cluster_version_if_none(self, v, client):
        current = SparkDiscovery.getOption(self.spark_cluster_version(), client=client)
        if current is None:
            create("map_node", self.spark_cluster_version().join(v),
                   recursive=True, ignore_existing=True, client=client)

    def conf(self):
        return self.discovery().join("conf")

    def master_wrapper(self):
        return self.discovery().join("master_wrapper")


class ClusterInfo(object):
    def __init__(self, master, shs, livy, workers, operation, workers_operation, multiop_mode, max_workers_count, user_slots):
        self.master = master
        self.shs = shs
        self.livy = livy
        self.workers = workers
        self.operation = operation
        self.workers_operation = workers_operation
        self.multiop_mode = multiop_mode
        self.max_workers_count = max_workers_count
        self.user_slots = user_slots

    def workers_count(self):
        return len(self.workers)


def cluster_info(yt_client, discovery_path):
    discovery = SparkDiscovery(discovery_path)
    master = SparkDiscovery.getOption(discovery.master_spark())
    shs = SparkDiscovery.getOption(discovery.shs())
    livy = SparkDiscovery.getOption(discovery.livy())
    operation = SparkDiscovery.getOption(discovery.operation())
    children = SparkDiscovery.getOptions(discovery.children_operations())
    if not children:
        worker_operation = operation
        multiop_mode = False
    else:
        worker_operation = children[0]
        multiop_mode = True
    yson_jobs = yt_client.list_jobs(worker_operation, job_state="running")
    jobs = yson_to_json(yson_jobs)['jobs']
    op = yson_to_json(yt_client.get_operation(worker_operation))
    resource_limits = op['runtime_parameters']['scheduling_options_per_pool_tree']['physical']['resource_limits']
    if 'user_slots' in resource_limits:
        user_slots = resource_limits['user_slots']
    else:
        user_slots = -1
    max_job_count = op['spec']['tasks']['workers']['job_count']

    def host(addr):
        h, _ = addr.split(':')
        return h

    workers = [j['address'] for j in jobs
               if 'address' in j and host(j['address']) != host(master) and host(j['address']) != host(shs)]
    return ClusterInfo(master, shs, livy, workers, operation, worker_operation, multiop_mode, max_job_count, user_slots)


def parse_memory(memory):
    if isinstance(memory, int):
        return memory
    if memory is None:
        return None
    units = {"gb": 1024 * 1024 * 1024, "mb": 1024 * 1024, "kb": 1024, "bb": 1, "b": 1}
    m = re.match(r"(\d+)(.*)", memory)
    value = int(m.group(1))
    unit = m.group(2).lower().strip()
    if len(unit) <= 1:
        unit = unit + "b"
    return value * units[unit]


def format_memory(memory_bytes):
    if memory_bytes is None:
        return None
    units = {"gb": 1024 * 1024 * 1024, "mb": 1024 * 1024, "kb": 1024, "bb": 1, "b": 1}
    if memory_bytes % units["gb"] == 0:
        return "{}G".format(memory_bytes // units["gb"])
    if memory_bytes % units["mb"] == 0:
        return "{}M".format(memory_bytes // units["mb"])
    if memory_bytes % units["kb"] == 0:
        return "{}K".format(memory_bytes // units["kb"])
    return "{}B".format(memory_bytes)


def get_spark_master(discovery, rest, yt_client=None):
    master_path = discovery.master_rest() if rest else discovery.master_spark()
    master = SparkDiscovery.get(master_path, client=yt_client)
    return "spark://{0}".format(master)


def scala_buffer_to_list(buffer):
    return [buffer.apply(i) for i in range(buffer.length())]


def default_user(client=None):
    return os.getenv("YT_USER") or get_user_name(client=client)


def default_token():
    return get_token()


def base_spark_conf(client, discovery):
    yt_proxy = call_get_proxy_address_url(required=True, client=client)
    yt_user = get_user_name(client=client)
    spark_cluster_version = SparkDiscovery.get(discovery.spark_cluster_version(), client=client)
    spark_cluster_conf = discovery.conf()
    master_wrapper_url = SparkDiscovery.get(discovery.master_wrapper(), client=client)
    conf = {
        "spark.hadoop.yt.proxy": yt_proxy,
        "spark.hadoop.yt.user": yt_user,
        "spark.master.rest.enabled": "true",
        "spark.shuffle.service.enabled": "true",
        "spark.eventLog.dir": "ytEventLog:/{}".format(discovery.event_log_table()),
        "spark.yt.cluster.version": spark_cluster_version,
        "spark.yt.version": spark_cluster_version,
        "spark.base.discovery.path": discovery.base_discovery_path
    }

    if exists(spark_cluster_conf, client=client):
        conf["spark.yt.cluster.confPath"] = str(spark_cluster_conf)
    if master_wrapper_url:
        conf["spark.hadoop.yt.masterWrapper.url"] = master_wrapper_url
    return conf


def set_conf(conf, dict_conf):
    if dict_conf is not None:
        for (key, value) in dict_conf.items():
            conf.set(key, value)


def default_discovery_dir(client=None):
    return os.getenv("SPARK_YT_DISCOVERY_DIR") \
        or YPath("//tmp").join(default_user(client=client)).join("spark-tmp")


def default_proxy():
    return os.getenv("YT_PROXY")


def default_tvm_secret():
    return os.getenv("SPARK_TVM_SECRET")


def default_tvm_id():
    return os.getenv("SPARK_TVM_ID")


def get_default_arg_parser(**kwargs):
    parser = argparse.ArgumentParser(**kwargs)
    parser.add_argument("--id", required=False)
    parser.add_argument("--discovery-path", required=False)
    parser.add_argument("--discovery-dir", required=False)
    parser.add_argument("--proxy", required=False, default=default_proxy())
    return parser


def parse_args(parser=None, parser_arguments=None, raw_args=None, client=None):
    parser_arguments = parser_arguments or {}
    parser = parser or get_default_arg_parser(**parser_arguments)
    args, unknown_args = parser.parse_known_args(args=raw_args)
    args.discovery_path = args.discovery_path or args.discovery_dir or default_discovery_dir(client=client)
    return args, unknown_args


def show_unspecified_args(unknown_args):
    if unknown_args:
        print("Unspecified arguments are ignored: %s" % ', '.join(unknown_arg.split('=')[0] for unknown_arg in unknown_args))


# backward compatibility
def tuple(element_types):
    from .types import tuple_type
    return tuple_type(element_types)


def get_spyt_home():
    spyt_home = checked_extract_spyt()
    if spyt_home is not None:
        return spyt_home
    import spyt
    return spyt.__path__[0]


# COMPAT(alex-shishkin): Remove when nobody use this instead of get_spark_home
def spark_home():
    return get_spark_home()


def get_spark_home():
    spark_home = os.environ.get("SPARK_HOME")
    if spark_home is not None:
        return spark_home
    spark_home = checked_extract_spark()
    if spark_home is not None:
        return spark_home
    try:
        spark_home = subprocess.check_output("find_spark_home.py").strip().decode("utf-8")
        return spark_home
    except Exception:
        raise RuntimeError("Unable to find SPARK_HOME automatically from {}".format(os.path.realpath(__file__)))


def call_get_proxy_address_url(**kwargs):
    # COMPAT(atokarew): replace this with get_proxy_address_url when compatibility with python wrapper <= 0.13.4 and spark cluster >= 1.73.0 is not important.
    import yt.wrapper.http_helpers

    if hasattr(yt.wrapper.http_helpers, 'get_proxy_address_url'):
        result = yt.wrapper.http_helpers.get_proxy_address_url(**kwargs)
    else:
        result = yt.wrapper.http_helpers.get_proxy_url(**kwargs)
    if result.startswith('http://'):
        result = result[7:]
    return result


def parse_bool(flag):
    return flag is not None and flag.lower() == 'true'


def _add_conf(spark_conf, spark_args):
    if spark_conf:
        for k, v in spark_conf.items():
            spark_args.append("--conf")
            spark_args.append("{}={}".format(k, v))


def check_spark_version(less_than=None, greater_than_or_equal=None):
    result = True
    current_version = Version(spark_version)
    if less_than:
        result = result and current_version < Version(less_than)
    if greater_than_or_equal:
        result = result and current_version >= Version(greater_than_or_equal)
    return result


def get_spyt_conf_dir():
    return os.path.join(get_spyt_home(), "conf")



import dis
import sys
import types

from pyspark.cloudpickle.cloudpickle import _extract_code_globals_cache, GLOBAL_OPS
from pyspark.sql.column import _to_java_column
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.types import StructType, StructField
from pyspark.sql import SparkSession

from .utils import check_spark_version


# DataFrameReader extensions
def read_yt(self, *paths):
    return self.format("yt").load(path=list(paths))


def _dict_to_struct(dict_type):
    return StructType([StructField(name, data_type) for (name, data_type) in dict_type.items()])


def read_schema_hint(self, fields):
    spark = SparkSession.builder.getOrCreate()
    struct_fields = []
    for name, data_type in fields.items():
        if isinstance(data_type, dict):
            data_type = _dict_to_struct(data_type)
        field = StructField(name, data_type)
        struct_fields.append(field)
    schema = StructType(struct_fields)

    jschema = spark._jsparkSession.parseDataType(schema.json())
    self._jreader = spark._jvm.tech.ytsaurus.spyt.PythonUtils.schemaHint(self._jreader, jschema)
    return self


# DataFrameWriter extensions
def write_yt(self, path, mode=None):
    def fix_path(path):
        return path[1:] if path.startswith("//") else path
    self.mode(mode)
    self.format("yt").save(fix_path(path))


def sorted_by(self, *cols, **kwargs):
    unique_keys = kwargs.get("unique_keys") or False
    import json
    return self.option("sort_columns", json.dumps(cols)).option("unique_keys", str(unique_keys))


def optimize_for(self, optimize_mode):
    return self.option("optimize_for", optimize_mode)


def write_schema_hint(self, fields):
    spark = SparkSession.builder.getOrCreate()

    jschema = spark._jvm.java.util.HashMap()
    for key, value in fields.items():
        jschema.put(key, value)
    self._jwrite = spark._jvm.tech.ytsaurus.spyt.PythonUtils.schemaHint(self._jwrite, jschema)
    return self


# DataFrame extensions
def withYsonColumn(self, colName, col):
    java_column = _to_java_column(col)
    return DataFrame(
        self._sc._jvm.tech.ytsaurus.spyt.PythonUtils.withYsonColumn(self._jdf, colName, java_column),
        self.sql_ctx
    )


def transform(self, func):
    return func(self)


if check_spark_version(less_than="3.4.0"):
    def _extract_code_globals(co):
        """
        Find all globals names read or written to by codeblock co
        """
        out_names = _extract_code_globals_cache.get(co)
        if out_names is None:
            # SPYT-415: From Spark 3.4
            if sys.version_info < (3, 11):
                names = co.co_names
                out_names = {names[oparg] for _, oparg in _walk_global_ops(co)}
            else:
                # We use a dict with None values instead of a set to get a
                # deterministic order (assuming Python 3.6+) and avoid introducing
                # non-deterministic pickle bytes as a results.
                out_names = {name: None for name in _walk_global_ops(co)}

            # Declaring a function inside another one using the "def ..."
            # syntax generates a constant code object corresonding to the one
            # of the nested function's As the nested function may itself need
            # global variables, we need to introspect its code, extract its
            # globals, (look for code object in it's co_consts attribute..) and
            # add the result to code_globals
            if co.co_consts:
                for const in co.co_consts:
                    if isinstance(const, types.CodeType):
                        out_names.update(_extract_code_globals(const))

            _extract_code_globals_cache[co] = out_names

        return out_names

    # SPYT-415: From Spark 3.4
    if sys.version_info < (3, 11):
        def _walk_global_ops(code):
            """
            Yield (opcode, argument number) tuples for all
            global-referencing instructions in *code*.
            """
            for instr in dis.get_instructions(code):
                op = instr.opcode
                if op in GLOBAL_OPS:
                    yield op, instr.arg
    else:
        def _walk_global_ops(code):
            """
            Yield referenced name for all global-referencing instructions in *code*.
            """
            for instr in dis.get_instructions(code):
                op = instr.opcode
                if op in GLOBAL_OPS:
                    yield instr.argval

    def _code_reduce(obj):
        """codeobject reducer"""
        # SPYT-415: From Spark 3.4
        # If you are not sure about the order of arguments, take a look at help
        # of the specific type from types, for example:
        # >>> from types import CodeType
        # >>> help(CodeType)
        if hasattr(obj, "co_exceptiontable"):  # pragma: no branch
            # Python 3.11 and later: there are some new attributes
            # related to the enhanced exceptions.
            args = (
                obj.co_argcount, obj.co_posonlyargcount,
                obj.co_kwonlyargcount, obj.co_nlocals, obj.co_stacksize,
                obj.co_flags, obj.co_code, obj.co_consts, obj.co_names,
                obj.co_varnames, obj.co_filename, obj.co_name, obj.co_qualname,
                obj.co_firstlineno, obj.co_linetable, obj.co_exceptiontable,
                obj.co_freevars, obj.co_cellvars,
            )
        elif hasattr(obj, "co_linetable"):  # pragma: no branch
            # Python 3.10 and later: obj.co_lnotab is deprecated and constructor
            # expects obj.co_linetable instead.
            args = (
                obj.co_argcount, obj.co_posonlyargcount,
                obj.co_kwonlyargcount, obj.co_nlocals, obj.co_stacksize,
                obj.co_flags, obj.co_code, obj.co_consts, obj.co_names,
                obj.co_varnames, obj.co_filename, obj.co_name,
                obj.co_firstlineno, obj.co_linetable, obj.co_freevars,
                obj.co_cellvars
            )
        elif hasattr(obj, "co_nmeta"):  # pragma: no cover
            # "nogil" Python: modified attributes from 3.9
            args = (
                obj.co_argcount, obj.co_posonlyargcount,
                obj.co_kwonlyargcount, obj.co_nlocals, obj.co_framesize,
                obj.co_ndefaultargs, obj.co_nmeta,
                obj.co_flags, obj.co_code, obj.co_consts,
                obj.co_varnames, obj.co_filename, obj.co_name,
                obj.co_firstlineno, obj.co_lnotab, obj.co_exc_handlers,
                obj.co_jump_table, obj.co_freevars, obj.co_cellvars,
                obj.co_free2reg, obj.co_cell2reg
            )
        elif hasattr(obj, "co_posonlyargcount"):
            # Backward compat for 3.9 and older
            args = (
                obj.co_argcount, obj.co_posonlyargcount,
                obj.co_kwonlyargcount, obj.co_nlocals, obj.co_stacksize,
                obj.co_flags, obj.co_code, obj.co_consts, obj.co_names,
                obj.co_varnames, obj.co_filename, obj.co_name,
                obj.co_firstlineno, obj.co_lnotab, obj.co_freevars,
                obj.co_cellvars
            )
        else:
            # Backward compat for even older versions of Python
            args = (
                obj.co_argcount, obj.co_kwonlyargcount, obj.co_nlocals,
                obj.co_stacksize, obj.co_flags, obj.co_code, obj.co_consts,
                obj.co_names, obj.co_varnames, obj.co_filename,
                obj.co_name, obj.co_firstlineno, obj.co_lnotab,
                obj.co_freevars, obj.co_cellvars
            )
        return types.CodeType, args



from spyt.dependency_utils import require_yt_client
require_yt_client()

from yt.wrapper.cli_helpers import ParseStructuredArgument  # noqa: E402
from spyt.enabler import SpytEnablers  # noqa: E402
from spyt.spec import SparkDefaultArguments  # noqa: E402


def add_parser_group(parser, enabler_arg, disabler_arg, dest_arg, default_value):
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument(enabler_arg, dest=dest_arg, action='store_true')
    group.add_argument(disabler_arg, dest=dest_arg, action='store_false')
    parser.set_defaults(**{dest_arg: default_value})


def add_default_launch_options(parser):
    parser.add_argument("--pool", required=False)
    parser.add_argument("--operation-alias", required=False)
    parser.add_argument("--network-project", required=False)
    parser.add_argument("--params", required=False, action=ParseStructuredArgument, dest="params",
                        default=SparkDefaultArguments.get_params())
    parser.add_argument("--spyt-version", required=False)
    parser.add_argument("--preemption_mode", required=False, default="normal")
    parser.add_argument("--cluster-log-level", required=False, default="INFO")

    default_enablers = SpytEnablers()
    add_parser_group(parser, '--enable-mtn', '--disable-mtn', 'enable_mtn', default_enablers.enable_mtn)
    add_parser_group(parser, '--prefer-ipv6', '--prefer-ipv4', 'enable_preference_ipv6', None)

    add_parser_group(parser, '--enable-tmpfs', '--disable-tmpfs', 'enable_tmpfs', False)
    add_parser_group(parser, '--enable-stderr-table', '--disable-stderr-table', 'enable_stderr_table', False)

    add_parser_group(parser, '--enable-tcp-proxy', '--disable-tcp-proxy', 'enable_tcp_proxy',
                     default_enablers.enable_tcp_proxy)
    parser.add_argument('--tcp-proxy-range-start', required=False, default=30000, type=int)
    parser.add_argument('--tcp-proxy-range-size', required=False, default=100, type=int)

    add_parser_group(parser, '--enable-rpc-job-proxy', '--disable-rpc-job-proxy', 'rpc_job_proxy', True)
    parser.add_argument("--rpc-job-proxy-thread-pool-size", required=False, default=4, type=int)

    parser.add_argument('--group-id', required=False, type=str)

    add_parser_group(parser, '--enable-squashfs', '--disable-squashfs', 'enable_squashfs',
                     default_enablers.enable_squashfs)
    parser.add_argument('--cluster-java-home', required=False, type=str)


def add_livy_options(parser):
    parser.add_argument('--livy-driver-cores', required=False,
                        default=SparkDefaultArguments.LIVY_DRIVER_CORES, type=int)
    parser.add_argument('--livy-driver-memory', required=False, default=SparkDefaultArguments.LIVY_DRIVER_MEMORY)
    parser.add_argument('--livy-max-sessions', required=False,
                        default=SparkDefaultArguments.LIVY_MAX_SESSIONS, type=int)


def add_hs_options(parser):
    parser.add_argument("--history-server-memory-limit",
                        required=False, default=SparkDefaultArguments.SPARK_HISTORY_SERVER_MEMORY_LIMIT)
    parser.add_argument("--history-server-memory-overhead",
                        required=False, default=SparkDefaultArguments.SPARK_HISTORY_SERVER_MEMORY_OVERHEAD)
    parser.add_argument("--history-server-cpu-limit",
                        required=False, default=SparkDefaultArguments.SPARK_HISTORY_SERVER_CPU_LIMIT, type=int)
    parser.add_argument("--shs-location", required=False)
    add_parser_group(parser, '--enable-advanced-event-log', '--disable-advanced-event-log', 'advanced_event_log', True)



import logging
import os

from spyt.dependency_utils import require_yt_client
require_yt_client()

from yt.wrapper import get, YPath, list as yt_list, exists  # noqa: E402
from yt.wrapper.common import update_inplace  # noqa: E402
from .utils import get_spyt_conf_dir  # noqa: E402
from .version import __scala_version__  # noqa: E402
from pyspark import __version__ as spark_version  # noqa: E402

SPARK_BASE_PATH = YPath(os.getenv('SPYT_ROOT_BASE_PATH', default="//home/spark"))

CONF_BASE_PATH = SPARK_BASE_PATH.join("conf")
GLOBAL_CONF_PATH = CONF_BASE_PATH.join("global")

SPYT_BASE_PATH = SPARK_BASE_PATH.join("spyt")
DISTRIB_BASE_PATH = SPARK_BASE_PATH.join("distrib")

RELEASES_SUBDIR = "releases"
PRE_RELEASES_SUBDIR = "pre-releases"
SNAPSHOTS_SUBDIR = "snapshots"

SELF_VERSION = __scala_version__

logger = logging.getLogger(__name__)


class SpytVersion:
    def __init__(self, version=None, major=0, minor=0, patch=0):
        if version is not None:
            self.major, self.minor, self.patch = map(int, version.split("-")[0].split("."))
        else:
            self.major, self.minor, self.patch = major, minor, patch

    def get_minor(self):
        return SpytVersion(major=self.major, minor=self.minor)

    def __gt__(self, other):
        return self.tuple() > other.tuple()

    def __ge__(self, other):
        return self.tuple() >= other.tuple()

    def __eq__(self, other):
        return self.tuple() == other.tuple()

    def __lt__(self, other):
        return self.tuple() < other.tuple()

    def __le__(self, other):
        return self.tuple() <= other.tuple()

    def tuple(self):
        return self.major, self.minor, self.patch

    def __str__(self):
        return f"{self.major}.{self.minor}.{self.patch}"


def validate_cluster_version(spark_cluster_version, client=None):
    if not check_cluster_version_exists(spark_cluster_version, client=client):
        raise RuntimeError("Unknown SPYT cluster version: {}. Available release versions are: {}".format(
            spark_cluster_version, get_available_cluster_versions(client=client)
        ))
    spyt_minor_version = SpytVersion(SELF_VERSION).get_minor()
    cluster_minor_version = SpytVersion(spark_cluster_version).get_minor()
    if spyt_minor_version < cluster_minor_version:
        logger.warning("You required SPYT version {} which is older than your local ytsaurus-spyt version {}."
                       "Please update your local ytsaurus-spyt".format(spark_cluster_version, SELF_VERSION))


def validate_versions_compatibility(spyt_version, spark_cluster_version):
    spyt_minor_version = SpytVersion(spyt_version).get_minor()
    spark_cluster_minor_version = SpytVersion(spark_cluster_version).get_minor()
    if spyt_minor_version > spark_cluster_minor_version:
        logger.warning("Your SPYT library has version {} which is older than your cluster version {}. "
                       "Some new features may not work as expected. "
                       "Please update your cluster with spark-launch-yt utility".format(spyt_version, spark_cluster_version))


def validate_mtn_config(enablers, network_project, tvm_id, tvm_secret):
    if enablers.enable_mtn and not network_project:
        raise RuntimeError("When using MTN, network_project arg must be set.")


def worker_num_limit(global_conf):
    return global_conf.get("worker_num_limit", 1000)


def cuda_toolkit_version(global_conf):
    return global_conf.get("cuda_toolkit_version", "11.0")


def validate_worker_num(worker_num, worker_num_lim):
    if worker_num > worker_num_lim:
        raise RuntimeError("Number of workers ({0}) is more than limit ({1})".format(worker_num, worker_num_lim))


def validate_ssd_config(disk_limit, disk_account):
    if disk_limit is not None and disk_account is None:
        raise RuntimeError("Disk account must be provided to use disk limit, please add --worker-disk-account option")


def get_available_cluster_versions(client=None):
    subdirs = yt_list(CONF_BASE_PATH.join(RELEASES_SUBDIR), client=client)
    return [x for x in subdirs if x != "spark-launch-conf"]


def check_cluster_version_exists(cluster_version, client=None):
    return exists(_get_version_conf_path(cluster_version), client=client)


def read_global_conf(client=None):
    return client.get(GLOBAL_CONF_PATH)


def read_remote_conf(global_conf, cluster_version, client=None):
    version_conf_path = _get_version_conf_path(cluster_version)
    version_conf = get(version_conf_path, client=client)
    version_conf["cluster_version"] = cluster_version
    return update_inplace(global_conf, version_conf)  # TODO(alex-shishkin): Might cause undefined behaviour


def read_metrics_conf():
    return read_conf("metrics.properties", delimiter="=")


def read_spark_defaults_conf():
    return read_conf("spark-defaults.conf")


def read_conf(filename, delimiter=None):
    props = {}
    try:
        conf_dir = get_spyt_conf_dir()
        filepath = os.path.join(conf_dir, filename)
        with open(filepath, 'r') as f:
            for line in f:
                line = line.strip()
                if line and not line.startswith('#'):
                    key_value = line.split(delimiter, 1)
                    if len(key_value) != 2:
                        continue
                    key, value = key_value
                    props[key.strip()] = value.strip()
    except Exception as e:
        print(f"Failed to load conf from {filename}: {e}")
    return props


def read_cluster_conf(path=None, client=None):
    if path is None:
        return {}
    return get(path, client=client)


def update_config_inplace(base, patch):
    file_paths = _get_or_else(patch, "file_paths", []) + _get_or_else(base, "file_paths", [])
    layer_paths = _get_or_else(patch, "layer_paths", []) + _get_or_else(base, "layer_paths", [])
    update_inplace(base, patch)
    base["file_paths"] = file_paths
    base["layer_paths"] = layer_paths
    return base


def validate_custom_params(params):
    if params and "enablers" in params:
        raise RuntimeError("Argument 'params' contains 'enablers' field, which is prohibited. "
                           "Use argument 'enablers' instead")


def get_available_spyt_versions(client=None):
    return yt_list(SPYT_BASE_PATH.join(RELEASES_SUBDIR), client=client)


def latest_ytserver_proxy_path(cluster_version, client=None):
    if cluster_version:
        return None
    global_conf = read_global_conf(client=client)
    symlink_path = global_conf.get("ytserver_proxy_path")
    if symlink_path is None:
        return None
    return get("{}&/@target_path".format(symlink_path), client=client)


def ytserver_proxy_attributes(path, client=None):
    return get("{}/@user_attributes".format(path), client=client)


def get_spark_distributive(client, enable_squashfs):
    distrib_root = DISTRIB_BASE_PATH.join(spark_version.replace('.', '/'))
    distrib_root_contents = yt_list(distrib_root, client=client)
    extension = '.squashfs' if enable_squashfs else '.tgz'
    spark_distributive = [x for x in distrib_root_contents if x.endswith(extension)]
    if len(spark_distributive) == 0:
        raise RuntimeError(f"Spark {spark_version} {extension} distributive doesn't exist "
                           f"at path {distrib_root} on cluster {client.config['proxy']['url']}")
    return (spark_distributive[0], distrib_root.join(spark_distributive[0]))


def _get_or_else(d, key, default):
    return d.get(key) or default


def _version_subdir(version):
    return SNAPSHOTS_SUBDIR if "SNAPSHOT" in version \
        else PRE_RELEASES_SUBDIR if any(r_type in version for r_type in ["alpha", "beta", "rc"]) \
        else RELEASES_SUBDIR


def _get_version_conf_path(cluster_version):
    return CONF_BASE_PATH.join(_version_subdir(cluster_version)).join(cluster_version).join("spark-launch-conf")



import glob
import logging
import os
import shutil
from tempfile import TemporaryDirectory, gettempdir
from typing import Optional
import zipfile
import tarfile

from spyt.dependency_utils import require_yt_client
require_yt_client()

from yt.wrapper.common import is_arcadia_python  # noqa: E402


logger = logging.getLogger(__name__)

extracted_spyt_dir: Optional[TemporaryDirectory] = None
extracted_spark_dir: Optional[TemporaryDirectory] = None

spark_distrib_dir = "yt/spark/spark-over-yt/spyt-package/src/main/spark/"


def _make_executables(directory):
    logger.debug(f"Making executables files in {directory}")
    for name in os.listdir(directory):
        os.chmod(os.path.join(directory, name), 0o755)


def _extract_resources(resource_key_prefix: str, remove_prefix: str, destination_dir: str):
    from library.python import resource
    if len(remove_prefix) > 0 and remove_prefix[-1] != '/':
        logger.warn("Last symbol in remove_prefix is not '/', you may try unpack to machine's root")
    logger.debug(f"Finding resource files with prefix {resource_key_prefix}")
    resource_paths = resource.resfs_files(resource_key_prefix)
    logger.debug(f"Found {len(resource_paths)} files. Extracting...")
    extracted_files = []
    for resource_path in resource_paths:
        assert resource_path.startswith(remove_prefix)
        relative_resource_path = resource_path[len(remove_prefix):]
        destination_path = os.path.join(destination_dir, relative_resource_path)
        os.makedirs(os.path.dirname(destination_path), exist_ok=True)
        with open(destination_path, 'bw') as out:
            out.write(resource.resfs_read(resource_path))
        extracted_files.append(destination_path)
    return extracted_files


def _spark_tar_members(tar_file, ):
    for member in tar_file.getmembers():
        slash_pos = member.path.find('/')
        if slash_pos > 0:
            member.path = member.path[slash_pos + 1:]
            yield member


# Return true if a PID found and  a process is alive
def _check_pid_in_dir(dir: str) -> bool:
    pid_file = os.path.join(dir, '.pid')
    if not os.path.exists(pid_file):
        return False
    try:
        with open(pid_file, 'r') as f:
            pid = int(f.read())
        os.kill(pid, 0)  # 0 - check signal. If it didn't fail (OSError) then the process is alive
        return True
    except Exception:
        return False


def _create_tmpdir_with_pid(prefix: str):
    temp_dir = TemporaryDirectory(prefix=prefix, ignore_cleanup_errors=True)
    with open(os.path.join(temp_dir.name, '.pid'), 'w') as f:
        f.write(str(os.getpid()))
    return temp_dir


def _extract_spark():
    temp_dir = _create_tmpdir_with_pid("spark_yamake_")
    logger.info(f"Created Spark temp dir {temp_dir}")
    files = _extract_resources(spark_distrib_dir, spark_distrib_dir, temp_dir.name)
    with tarfile.open(files[0], 'r:gz') as tar_file:
        tar_file.extractall(temp_dir.name, members=_spark_tar_members(tar_file), filter='tar')
    logger.info("Spark files extracted successfully")
    return temp_dir


def _extract_spyt():
    temp_dir = _create_tmpdir_with_pid("spyt_yamake_")
    logger.info(f"Created Spyt temp dir {temp_dir}")
    spyt_original_dir = "yt/spark/spark-over-yt/spyt-package/src/main/spyt_cluster/"
    build_dir = temp_dir.name
    files = _extract_resources(spyt_original_dir, spyt_original_dir, build_dir)
    with zipfile.ZipFile(files[0], 'r') as zip_ref:
        zip_ref.extractall(build_dir)
    package_dir = build_dir + "/spyt-package"
    for file_name in os.listdir(package_dir):
        os.rename(package_dir + "/" + file_name, build_dir + "/" + file_name)
    logger.info("Spyt files extracted successfully")
    return temp_dir


def checked_extract_spyt() -> Optional[str]:
    if not is_arcadia_python():
        return None

    global extracted_spyt_dir
    logger.debug("Arcadia python found")
    if not extracted_spyt_dir:
        extracted_spyt_dir = _extract_spyt()
    logger.debug(f"Current extracted Spyt location {extracted_spyt_dir}")
    return extracted_spyt_dir.name


def checked_extract_spark() -> Optional[str]:
    if not is_arcadia_python():
        return None

    global extracted_spark_dir
    logger.debug("Arcadia python found")
    if not extracted_spark_dir:
        extracted_spark_dir = _extract_spark()
        os.environ["SPARK_HOME"] = extracted_spark_dir.name
    logger.debug(f"Current extracted Spark location {extracted_spark_dir}")
    return extracted_spark_dir.name


def _remove_from_tempdir(prefix: str, only_dead: bool):
    tempdir = gettempdir()
    for old_tempdir in glob.glob(os.path.join(tempdir, prefix + "*")):
        if not only_dead or not _check_pid_in_dir(old_tempdir):
            logger.info(f"Removing temp dir {old_tempdir}")
            shutil.rmtree(old_tempdir)


def clean_extracted():
    global extracted_spark_dir, extracted_spyt_dir
    if extracted_spark_dir:
        extracted_spark_dir.cleanup()
        extracted_spark_dir = None
    if extracted_spyt_dir:
        extracted_spyt_dir.cleanup()
        extracted_spyt_dir = None


def remove_all_temp_files(only_dead=True):
    _remove_from_tempdir("spark_yamake_", only_dead)
    _remove_from_tempdir("spyt_yamake_", only_dead)
